/**
Copyright 2016 The Chromium Authors. All rights reserved.
Use of this source code is governed by a BSD-style license that can be
found in the LICENSE file.
**/

require("../../base/range.js");
require("../metric_registry.js");
require("./utils.js");
require("../../value/numeric.js");
require("../../value/unit.js");
require("../../value/value.js");

'use strict';

global.tr.exportTo('tr.metrics.v8', function() {
  // The time window size for mutator utilization computation.
  // It is equal to the duration of one frame corresponding to 60 FPS rendering.
  var TARGET_FPS = 60;
  var MS_PER_SECOND = 1000;
  var WINDOW_SIZE_MS = MS_PER_SECOND / TARGET_FPS;

  function gcMetric(values, model) {
    addDurationOfTopEvents(values, model);
    addTotalDurationOfTopEvents(values, model);
    addDurationOfSubEvents(values, model);
    addIdleTimesOfTopEvents(values, model);
    addTotalIdleTimesOfTopEvents(values, model);
    addV8ExecuteMutatorUtilization(values, model);
  }

  tr.metrics.MetricRegistry.register(gcMetric);

  var timeDurationInMs_smallerIsBetter =
      tr.v.Unit.byName.timeDurationInMs_smallerIsBetter;
  var percentage_biggerIsBetter =
      tr.v.Unit.byName.normalizedPercentage_biggerIsBetter;

  var numericBuilder = new tr.v.NumericBuilder(
      timeDurationInMs_smallerIsBetter, 0);
  // 0.1 steps from 0 to 20 since it is the most common range.
  numericBuilder.addLinearBins(20, 200);
  // Exponentially increasing steps from 20 to 200.
  numericBuilder.addExponentialBins(200, 100);

  function createNumericForTopEventTime() {
    var n = numericBuilder.build();
    n.customizeSummaryOptions({
        avg: true,
        count: true,
        max: true,
        min: false,
        std: true,
        sum: true,
        percentile: [0.90]});
    return n;
  }

  function createNumericForSubEventTime() {
    var n = numericBuilder.build();
    n.customizeSummaryOptions({
        avg: true,
        count: false,
        max: true,
        min: false,
        std: false,
        sum: false,
        percentile: [0.90]
    });
    return n;
  }

  function createNumericForIdleTime() {
    var n = numericBuilder.build();
    n.customizeSummaryOptions({
        avg: true,
        count: false,
        max: true,
        min: false,
        std: false,
        sum: true,
        percentile: []
    });
    return n;
  }

  function createPercentage(numerator, denominator) {
    var percentage = denominator === 0 ? 0 : numerator / denominator * 100;
    return new tr.v.ScalarNumeric(percentage_biggerIsBetter, percentage);
  }

  function isNotForcedTopGarbageCollectionEvent(event) {
    // We exclude garbage collection events forced by benchmark runner,
    // because they cannot happen in real world.
    return tr.metrics.v8.utils.isTopGarbageCollectionEvent(event) &&
           !tr.metrics.v8.utils.isForcedGarbageCollectionEvent(event);
  }

  function isNotForcedSubGarbageCollectionEvent(event) {
    // We exclude garbage collection events forced by benchmark runner,
    // because they cannot happen in real world.
    return tr.metrics.v8.utils.isSubGarbageCollectionEvent(event) &&
           !tr.metrics.v8.utils.isForcedGarbageCollectionEvent(event);
  }

  /**
   * Example output:
   * - v8-gc-full-mark-compactor.
   */
  function addDurationOfTopEvents(values, model) {
    tr.metrics.v8.utils.groupAndProcessEvents(model,
      isNotForcedTopGarbageCollectionEvent,
      tr.metrics.v8.utils.topGarbageCollectionEventName,
      function(name, events) {
        var cpuDuration = createNumericForTopEventTime();
        events.forEach(function(event) {
          cpuDuration.add(event.cpuDuration);
        });
        values.addValue(new tr.v.NumericValue(name, cpuDuration));
      }
    );
  }

  /**
   * Example output:
   * - v8-gc-total
   */
  function addTotalDurationOfTopEvents(values, model) {
    tr.metrics.v8.utils.groupAndProcessEvents(model,
      isNotForcedTopGarbageCollectionEvent,
      event => 'v8-gc-total',
      function(name, events) {
        var cpuDuration = createNumericForTopEventTime();
        events.forEach(function(event) {
          cpuDuration.add(event.cpuDuration);
        });
        values.addValue(new tr.v.NumericValue(name, cpuDuration));
      }
    );
  }

  /**
   * Example output:
   * - v8-gc-full-mark-compactor-evacuate.
   */
  function addDurationOfSubEvents(values, model) {
    tr.metrics.v8.utils.groupAndProcessEvents(model,
      isNotForcedSubGarbageCollectionEvent,
      tr.metrics.v8.utils.subGarbageCollectionEventName,
      function(name, events) {
        var cpuDuration = createNumericForSubEventTime();
        events.forEach(function(event) {
          cpuDuration.add(event.cpuDuration);
        });
        values.addValue(new tr.v.NumericValue(name, cpuDuration));
      }
    );
  }

  /**
   * Example output:
   * - v8-gc-full-mark-compactor_idle_deadline_overrun,
   * - v8-gc-full-mark-compactor_outside_idle,
   * - v8-gc-full-mark-compactor_percentage_idle.
   */
  function addIdleTimesOfTopEvents(values, model) {
    tr.metrics.v8.utils.groupAndProcessEvents(model,
      isNotForcedTopGarbageCollectionEvent,
      tr.metrics.v8.utils.topGarbageCollectionEventName,
      function(name, events) {
        addIdleTimes(values, model, name, events);
      }
    );
  }

  /**
   * Example output:
   * - v8-gc-total_idle_deadline_overrun,
   * - v8-gc-total_outside_idle,
   * - v8-gc-total_percentage_idle.
   */
  function addTotalIdleTimesOfTopEvents(values, model) {
    tr.metrics.v8.utils.groupAndProcessEvents(model,
      isNotForcedTopGarbageCollectionEvent,
      event => 'v8-gc-total',
      function(name, events) {
        addIdleTimes(values, model, name, events);
      }
    );
  }

  function addIdleTimes(values, model, name, events) {
    var cpuDuration = createNumericForIdleTime();
    var insideIdle = createNumericForIdleTime();
    var outsideIdle = createNumericForIdleTime();
    var idleDeadlineOverrun = createNumericForIdleTime();
    events.forEach(function(event) {
      var idleTask = tr.metrics.v8.utils.findParent(
          event, tr.metrics.v8.utils.isIdleTask);
      var inside = 0;
      var overrun = 0;
      if (idleTask) {
        var allottedTime = idleTask['args']['allotted_time_ms'];
        if (event.duration > allottedTime) {
          overrun = event.duration - allottedTime;
          // Don't count time over the deadline as being inside idle time.
          // Since the deadline should be relative to wall clock we
          // compare allotted_time_ms with wall duration instead of thread
          // duration, and then assume the thread duration was inside idle
          // for the same percentage of time.
          inside = event.cpuDuration * allottedTime / event.duration;
        } else {
          inside = event.cpuDuration;
        }
      }
      cpuDuration.add(event.cpuDuration);
      insideIdle.add(inside);
      outsideIdle.add(event.cpuDuration - inside);
      idleDeadlineOverrun.add(overrun);
    });
    values.addValue(new tr.v.NumericValue(
        name + '_idle_deadline_overrun',
        idleDeadlineOverrun));
    values.addValue(new tr.v.NumericValue(
        name + '_outside_idle', outsideIdle));
    var percentage = createPercentage(insideIdle.sum,
                                      cpuDuration.sum);
    values.addValue(new tr.v.NumericValue(
        name + '_percentage_idle', percentage));
  }

  function addV8ExecuteMutatorUtilization(values, model) {
    tr.metrics.v8.utils.groupAndProcessEvents(model,
        tr.metrics.v8.utils.isTopV8ExecuteEvent,
        event => 'v8-execute',
        function(name, events) {
          events.sort((a, b) => a.start - b.start);
          var time = 0;
          var pauses = [];
          // Glue together the v8.execute events and adjust the GC pause
          // times accordingly.
          for (var topEvent of events) {
            for (var e of topEvent.enumerateAllDescendents()) {
              if (isNotForcedTopGarbageCollectionEvent(e)) {
                pauses.push({ start: e.start - topEvent.start + time,
                              end: e.end - topEvent.start + time });
              }
            }
            time += topEvent.duration;
          }
          // Now we have one big v8.execute interval from 0 to |time| and
          // a list of GC pauses.
          var mutatorUtilization = tr.metrics.v8.utils.mutatorUtilization(
              0, time, WINDOW_SIZE_MS, pauses);
          [0.90, 0.95, 0.99].forEach(function(percent) {
            var value = new tr.v.ScalarNumeric(percentage_biggerIsBetter,
                mutatorUtilization.percentile(1.0 - percent) * 100);
            values.addValue(new tr.v.NumericValue(
                'v8-execute-mutator-utilization_pct_0' +
                percent * 100,
                value));
          });
          var value = new tr.v.ScalarNumeric(percentage_biggerIsBetter,
              mutatorUtilization.min);
          values.addValue(new tr.v.NumericValue(
              'v8-execute-mutator-utilization_min', value));
        }
    );
  }

  return {
    gcMetric: gcMetric,
    WINDOW_SIZE_MS: WINDOW_SIZE_MS // For testing purposes only.
  };
});
